"""Simplified DOI interface."""
import json
import re
import string
import logging
from typing import Optional, Dict
import requests
from .dtos import PaperDTO
from . import config


class Doi:
    """Interface w/ the doi.org api."""

    _cache: Dict[str, PaperDTO] = {}

    def get_doi_link(self: 'Doi', doi: str) -> str:
        """Assemble doi link."""
        return f"http://doi.org/{doi}"

    def load_doi_data(self: 'Doi', doi: str) -> str:
        """Load data for doi."""
        headers = {
            'Accept': 'application/json',
        }
        return requests.get(f'http://dx.doi.org/{doi}',
                            headers=headers).text

    def parse_doi_json(self: 'Doi', jsoncontent: str) -> PaperDTO:
        """Tranform doi json to PaperDTO."""
        info = json.loads(jsoncontent)

        if config.debug:
            filename = format_filename(f"doi-{info['DOI']}.json")
            with open(filename, "w") as file:
                json.dump(info, file, indent=4)

            if 'author' in info and info['author'][0]['sequence'] != 'first':
                logging.error("incorrect first author %s", jsoncontent)

        author = (f"{info['author'][0]['given']} {info['author'][0]['family']}"
                  if "author" in info
                  else "Author N/A")

        authors = (", ".join([(f"{a['given'] if 'given' in a else ''} "
                               f"{a['family'] if 'family' in a else ''} "
                               f"{a['name'] if 'name' in a else ''}").strip()
                              for a in info['author']])
                   if "author" in info
                   else "Authors N/A")

        title = (info['title']
                 if "title" in info and isinstance(info['title'], str)
                 else "Title N/A")
        journal = (info['publisher']
                   if "publisher" in info
                   else "Journal N/A")
        year = info['created']['date-parts'][0][0]
        doi = info['DOI']
        abstract = info['abstract'] if "abstract" in info else "Abstract N/A"

        slug = (f"{info['author'][0]['family']}{year}"
                if "author" in info else "N/A")

        return PaperDTO(author, authors, title, journal, year, abstract, doi,
                        slug)

    def get_bibtex(self: 'Doi', doi: str) -> str:
        """Get bibtex string for doi.

        Try also:
        curl -LH "Accept: text/bibliography; style=bibtex" \
                    http://dx.doi.org/10.1007/s00445-020-01390-8

        """
        headers = {
            'Accept': 'text/bibliography; style=bibtex',
        }
        req = requests.get(f'http://dx.doi.org/{doi}', headers=headers)

        # guess and set correct encoding
        req.encoding = "utf-8"

        return req.text

    def get_info(self: 'Doi', doi: str) -> Optional[PaperDTO]:
        """Get information for doi."""
        if doi in self._cache:
            return self._cache[doi]
        try:
            jsoncontent = self.load_doi_data(doi)
            data = self.parse_doi_json(jsoncontent)
            self._cache[doi] = data
            return data
        except json.decoder.JSONDecodeError:
            if config.debug:
                with open(format_filename(doi+"-broken.txt"), "w") as file:
                    file.write(str(jsoncontent))

            return None

    def extract_doi(self: 'Doi', hay: str) -> Optional[str]:
        """Parse doi from string, or None if not found.

        >>> Doi().extract_doi("https://doi.org/10.1093/petrology/egaa077")
        '10.1093/petrology/egaa077'
        """
        pattern = r'\b10\.\d{4,9}/[-._;()/:A-Z0-9]+'
        matches = re.compile(pattern, re.I).search(hay)
        return matches.group() if matches else None


def format_filename(inputstr: str) -> str:
    """Take a string and return a valid filename constructed from the string.

    Uses a whitelist approach: any characters not present in valid_chars are
    removed. Also spaces are replaced with underscores.

    Note: this method may produce invalid filenames such as ``, `.` or `..`
    When I use this method I prepend a date string like '2009_01_15_19_46_32_'
    and append a file extension like '.txt', so I avoid the potential of using
    an invalid filename.

    https://gist.github.com/seanh/93666
    """
    valid_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)
    filename = ''.join(c for c in inputstr if c in valid_chars)
    filename = filename.replace(' ', '_')
    return filename



"""Text user interface main."""
from functools import partial
import sys
import logging
import asyncio
import webbrowser
from typing import Any, Dict
import urwid  # type: ignore
from ..papersurfer import Papersurfer
from ..exceptions import ConfigError
from ..doi import Doi
from ..dtos import PostDTO, PaperDTO
from .postdialog import PostDialog
from .prettybutton import PrettyButton
from .paperdetail import PaperDetail
from .highlight import highlight


class PapersurferUi:
    """Provide UI and interface with mattermost class."""

    _palette = [
        ('button', 'default,bold', 'default'),
        ('I say', 'default,bold', 'default', 'bold'),
        ('needle', 'default, bold, underline', 'default', 'bold'),
        ('highlight', 'black', 'dark blue'),
        ('banner', 'black', 'light gray'),
        ('selectable', 'white', 'black'),
        ('focus', 'black', 'light gray'),
        ('papertitle', 'default,bold', 'default', 'bold')
    ]

    def __init__(self: "PapersurferUi", url: str, channel: str, username: str,
                 password: str) -> None:
        self.papersurfer = Papersurfer(url, channel, username, password)

        # TODO: move to Papersurfer or cache DOI.getInfo
        self.papers: Dict[str, PaperDTO] = {}
        self._screen = urwid.raw_display.Screen()

        self.size = self._screen.get_cols_rows()

        ask = urwid.Edit(('I say', u"Filter?\n"))
        exitbutton = PrettyButton(u'Exit', on_press=self.on_exit_clicked)
        self.exportbutton = PrettyButton(u'Export filtered list as bibtex',
                                         on_press=self.on_export_clicked)
        submitbutton = PrettyButton('Submit paper',
                                    on_press=self.open_submit_paper)
        div = urwid.Divider(u'-')

        body = [urwid.Text("")]
        self.listcontent = urwid.SimpleFocusListWalker(body)
        self.post_list = urwid.ListBox(self.listcontent)
        paperlist = urwid.BoxAdapter(self.post_list, self.size[1] - 5)

        urwid.connect_signal(self.listcontent, "modified", self._h_list_focus)

        buttonrow = urwid.Columns([exitbutton, self.exportbutton,
                                   submitbutton])
        self.paperdetail = PaperDetail()
        self.pile = urwid.Pile([ask,
                                div,
                                # TODO: horizontal or vertical
                                urwid.Columns([paperlist,
                                               self.paperdetail.render()]),

                                div,
                                buttonrow])
        self.top = urwid.Filler(self.pile, valign='middle')
        self._over = urwid.Overlay(
            self.loading_indicator(),
            self.top,
            align='center',
            valign='middle',
            width=20,
            height=10
        )

        urwid.connect_signal(ask, 'change', self.onchange)

        self.loop = asyncio.get_event_loop()
        evl = urwid.AsyncioEventLoop(loop=self.loop)
        self.mainloop = urwid.MainLoop(self._over, self._palette,
                                       unhandled_input=self._h_unhandled_input,
                                       event_loop=evl)

        self.loop.run_in_executor(None, self._load_list)
        self.loop.run_in_executor(None, self._update_data)

        self.mainloop.run()
        try:
            self.papersurfer.mattermost.login()
        except ConfigError:
            sys.exit(1)

    def _h_list_focus(self: "PapersurferUi") -> None:
        post_idx = self.post_list.get_focus()[1]
        if post_idx is not None:
            post = self.papersurfer.get_posts_filtered()[post_idx]
            self.loop.run_in_executor(None, self._load_doi, post.doi)
        else:
            post = PostDTO()
        self.paperdetail.post = post

    def _load_doi(self: "PapersurferUi", doi: str) -> None:
        if doi not in self.papers:
            self.papers[doi] = Doi().get_info(doi) or PaperDTO()

        # kick paperdetail to let it know new data is available
        self.paperdetail.papers = self.papers

    def _h_unhandled_input(self: "PapersurferUi", key: str) -> None:
        """Handle keyboard input not otherwise handled."""
        if key == "esc":
            raise urwid.ExitMainLoop()

    def _h_load_list(self: "PapersurferUi", _loop: Any, _data: Any) -> None:
        """Handle load list alarm."""
        self._load_list()

    def _load_list(self: "PapersurferUi") -> None:
        """Load and display paper list."""
        posts = self.papersurfer.get_posts()
        if posts:
            body = [self.list_item(post) for post in posts]
            self.listcontent.clear()
            self.listcontent.extend(body)
            self.mainloop.widget = self.top

    def _h_update_data(self: "PapersurferUi", _loop: Any, _data: Any) -> None:
        """Handle update data alarm."""
        self._update_data()

    def _update_data(self: "PapersurferUi") -> None:
        """Load and display paper list."""
        try:
            self.papersurfer.load()
        except ConfigError:
            logging.error("Failed to retrieve data from mattermost.")
            raise urwid.ExitMainLoop() from ConfigError
        self._load_list()

    def loading_indicator(self: "PapersurferUi") -> Any:
        """Create loading indicator dialog."""
        body_text = urwid.Text("Loading...", align='center')
        body_filler = urwid.Filler(body_text, valign='middle')
        body_padding = urwid.Padding(
            body_filler,
            left=1,
            right=1
        )

        return urwid.Frame(body_padding)

    def list_item(self: "PapersurferUi", post: PostDTO,
                  needle: str = "") -> Any:
        """Render a post in the list of posts."""
        text = f"{post.message} ({post.reporter})"
        title = urwid.Text(highlight(text, needle))

        discuss_button = PrettyButton("Open Discussion",
                                      on_press=partial(self.h_open_discussion,
                                                       post))
        doi_button = PrettyButton("Open DOI",
                                  on_press=partial(self.h_open_doi, post))

        button_bar = urwid.Columns([discuss_button, doi_button])
        pile = urwid.Pile([title, button_bar, urwid.Divider()])
        return pile

    def updscrn(self: "PapersurferUi") -> None:
        """Update (redraw) screen."""
        self.mainloop.draw_screen()

    def onchange(self: "PapersurferUi", _: Any, needle: str) -> None:
        """Handle filter change."""
        self.papersurfer.searchstring = needle
        self.paperdetail.searchstring = needle
        self.listcontent.clear()
        self.listcontent.extend([
            self.list_item(paper, needle)
            for paper in self.papersurfer.get_posts_filtered()])

    def running_export(self: "PapersurferUi", state: bool) -> None:
        """Set exporting state."""
        label = self.exportbutton.get_label()
        running_indicator = " (running...)"
        if state:
            self.exportbutton.set_label(label + running_indicator)
        else:
            self.exportbutton.set_label(label.replace(running_indicator, ""))
        self.updscrn()

    def on_exit_clicked(self: "PapersurferUi", _: Any) -> None:
        """Handle exitbutton click and exit."""
        raise urwid.ExitMainLoop()

    def on_export_clicked(self: "PapersurferUi", _: Any) -> None:
        """Handle exitbutton click and exit."""
        self.running_export(True)
        self.papersurfer.export_to_bibtex()
        self.running_export(False)

    def h_open_discussion(self: "PapersurferUi", post: PostDTO,
                          _: Any) -> None:
        """Handle click/enter on discussion button."""
        self.open_discussion(post)

    def h_open_doi(self: "PapersurferUi", post: PostDTO, _: Any) -> None:
        """Handle click/enter on doi button."""
        self.open_doi(post)

    def open_discussion(self: "PapersurferUi", post: PostDTO) -> None:
        """Open Mattermost post in browser."""
        webbrowser.open(post.link)

    def open_doi(self: "PapersurferUi", post: PostDTO) -> None:
        """Open paper page in browser."""
        webbrowser.open(Doi().get_doi_link(post.doi))

    def h_close_dialog(self: "PapersurferUi", _: Any) -> None:
        """Handle close dialog button."""
        self.close_dialog()

    def close_dialog(self: "PapersurferUi") -> None:
        """Close currently open dialog."""
        self.mainloop.widget = self.top

    def h_close_submit_paper(self: "PapersurferUi", _: Any) -> None:
        """Reload fresh data when closing submit paper dialog."""
        self._update_data()
        self.close_dialog()

    def open_submit_paper(self: "PapersurferUi", _: Any) -> None:
        """Open submit paper dialog."""
        self._over = urwid.Overlay(
            PostDialog(self.papersurfer.mattermost,
                       self.h_close_submit_paper, self.mainloop),
            self.top,
            align='center',
            valign='middle',
            width=100,
            height=200
        )

        self.mainloop.widget = self._over



"""Element to display paper details."""
from typing import Dict
import urwid  # type: ignore
from ..dtos import PaperDTO, PostDTO
from .highlight import highlight


class PaperDetail:
    """Render details to a post, including async paper data."""
    def __init__(self: "PaperDetail") -> None:
        self._papers: Dict[str, PaperDTO] = {}
        self._post = PostDTO()
        self._paper = PaperDTO()

        self.keywords = urwid.Text("")
        self.summary = urwid.Text("")
        self.paper_title = urwid.Text("")
        self.paper_authors = urwid.Text("")
        self.paper_journal = urwid.Text("")
        self.paper_doi = urwid.Text("")
        self.paper_abstract = urwid.Text("")
        self.bibtex = urwid.Text("")

#        self.updatecounter = 0
#        self.updates = urwid.Text("")

        self._searchstring = ""

    @property
    def post(self: "PaperDetail") -> PostDTO:
        """Get post."""
        return self._post

    @post.setter
    def post(self: "PaperDetail", post: PostDTO) -> None:
        """Set post, reload paper, re-render."""
        self._post = post
        self._paper = (self._papers[post.doi]
                       if post.doi in self._papers
                       else PaperDTO())
        self._update()

    @property
    def papers(self: "PaperDetail") -> Dict[str, PaperDTO]:
        """Get papers."""
        return self._papers

    @papers.setter
    def papers(self: "PaperDetail", papers: Dict[str, PaperDTO]) -> None:
        """Set papers and re-render if applicable."""
        self._papers = papers
        self._paper = (self._papers[self._post.doi]
                       if self._post.doi in self._papers
                       else PaperDTO())
        self._update()

    @property
    def searchstring(self: "PaperDetail") -> str:
        """Get searchstring."""
        return self._searchstring

    @searchstring.setter
    def searchstring(self: "PaperDetail", searchstring: str) -> None:
        """Set searchstring and update."""
        self._searchstring = searchstring
        self._update()

    def _update(self: "PaperDetail") -> None:
        """Re-render content."""
        self.summary.set_text(highlight(self._post.summary, self.searchstring))
        self.keywords.set_text(
            highlight(", ".join(self._post.keywords), self.searchstring))
        self.paper_title.set_text(
            highlight(self._paper.title, self.searchstring))
        self.paper_authors.set_text(
            highlight(self._paper.authors, self.searchstring))
        self.paper_journal.set_text(
            highlight(self._paper.journal, self.searchstring))
        self.paper_doi.set_text(highlight(self._paper.doi, self.searchstring))
        self.paper_abstract.set_text(
            highlight(self._paper.abstract, self.searchstring))
#        self.updatecounter += 1
#        self.updates.set_text("Updates: " + str(self.updatecounter))

    def render(self: "PaperDetail") -> urwid.Pile:
        """Create Dialog with paper details."""
        body_pile = urwid.Pile([
            self.paper_title,
            self.paper_authors,
            self.paper_journal,
            self.paper_doi,
            self.paper_abstract,
            self.summary,
            self.bibtex,
            self.keywords,
            # self.updates
        ])

        return body_pile



"""Simplified mattermost interface."""
import logging
import time
import json
import re
from typing import Optional, List, Dict, Any
import requests
import mattermostdriver  # type: ignore
from .exceptions import ConfigError
from .dtos import PostDTO
from .doi import Doi
from . import config


def extract_by_keyword(keyword: str, text: str, indicators: List[str]) -> str:
    """Extract content by keyword.

    >>> text = "Keywords: keyword1, keyword2"
    >>> extract_by_keyword("Keywords:", text, ["Keyword"])
    ' keyword1, keyword2'
    >>> text = "Keywords: keyword1, keyword2"
    >>> extract_by_keyword("Keywords:", text, [])
    ' keyword1, keyword2'
    """
    pattern = '{}(.*)((?:\n.*?)+?)?({}|$)'
    try:
        expression = pattern.format(keyword, "|".join(indicators))
        matches = re.findall(expression, text, flags=re.IGNORECASE)
        return "".join(matches[0][0:-1])
    except IndexError:
        return ""


def extract_summary_keywords(text: str) -> Dict[str, Any]:
    r"""Parse keywords and summary from post messages.

    >>> text = "Keywords: keyword1, keyword2"
    >>> extract_summary_keywords(text)
    {'keywords': ['keyword1', 'keyword2'], 'summary': '', 'message': ''}

    >>> text = "Keywords: keyword1; keyword2"
    >>> extract_summary_keywords(text)
    {'keywords': ['keyword1', 'keyword2'], 'summary': '', 'message': ''}

    >>> text = "Keywords: \n keyword1 \n keyword2"
    >>> extract_summary_keywords(text)
    {'keywords': ['keyword1', 'keyword2'], 'summary': '', 'message': ''}

    >>> text = "Summary: Test summary"
    >>> extract_summary_keywords(text)
    {'keywords': [], 'summary': 'Test summary', 'message': ''}

    >>> text = "Key points: Test key points"
    >>> extract_summary_keywords(text)
    {'keywords': [], 'summary': 'Test key points', 'message': ''}

    >>> text = '''Some Paper
    ... DOI: 1234.12/12
    ...
    ... keywords: Stu**[a-n]ff, Interest$
    ...
    ... summary:
    ... Interesting stuff
    ... more'''
    >>> r = extract_summary_keywords(text)  # doctest: +ELLIPSIS
    >>> r['keywords']
    ['Stu**[a-n]ff', 'Interest$']
    >>> r['summary']
    'Interesting stuff\nmore'
    >>> r['message']
    'Some Paper\nDOI: 1234.12/12'

    >>> text = '''Some Paper
    ... Key Points
    ... Interesting stuff'''
    >>> extract_summary_keywords(text)
    {'keywords': [], 'summary': 'Interesting stuff', 'message': 'Some Paper'}

    # TODO:
    # >>> text = '''Storage and Evolution
    # ... Jacob D. Klug [Klug2020]
    # ... https://doi.org/10.1029/2020JB019475
    # ...     Magma storage
    # ...     Mafic recharge
    # ... '''
    # >>> r = extract_summary_keywords(text)
    # >>> r['message']
    # ''
    """
    # some phrases that indicate a summary or keyword section
    summaryindicator = ["key points:", "key points\n", "keypoints:",
                        "keypoints\n", "summary:", "summary\n"]
    keywordindicator = ["keywords:"]
    indicators = summaryindicator + keywordindicator

    kwstrs = ""
    for keyword in keywordindicator:
        kwstr = extract_by_keyword(keyword, text, indicators)
        textsearch = re.compile(re.escape(keyword + kwstr), re.IGNORECASE)
        text = textsearch.sub('', text)
        kwstrs += kwstr

    keywords = [kw.strip() for kw in re.split(',|;|\n', kwstrs.strip()) if kw]

    summary = ""
    for keyword in summaryindicator:
        kwstr = extract_by_keyword(keyword, text, indicators)
        textsearch = re.compile(keyword + kwstr, re.IGNORECASE)
        text = textsearch.sub('', text)
        summary += kwstr

    return {"keywords": keywords,
            "summary": summary.strip(),
            "message": text.strip()}


class Mattermost:
    """Provide a simplified interaction w/ mattermost api."""
    def __init__(self: 'Mattermost', url: str, channelname: str, username: str,
                 password: str) -> None:
        self.posts: List[PostDTO] = []
        self._mattermost = mattermostdriver.Driver({
            'url': url,
            'login_id': username,
            'password': password,
            'port': 443
        })

        self.teamname = ""
        self._url = url
        self._loggedin = False
        self._reporters: Dict[str, str] = {}
        self._channelname = channelname
        self._channel: Optional[str] = None

    def login(self: 'Mattermost') -> None:
        """Log into mattermost and raise ConfigError on fail."""
        try:
            self._mattermost.login()
        except requests.exceptions.ConnectionError:
            print("Mattermost url is incorrect.")
            raise ConfigError from requests.exceptions.ConnectionError
        except (mattermostdriver.exceptions.NoAccessTokenProvided,
                requests.exceptions.InvalidURL,
                requests.exceptions.HTTPError) as exception:
            print("Failed to log into Mattermost.")
            raise ConfigError from exception

        try:
            self._channel = self._get_channel(self._channelname)
        except ConfigError:
            print("Couldn't find Mattermost channel.")
            raise ConfigError from ConfigError

        self._loggedin = True

    def _get_channel(self: "Mattermost", channelname: str) -> Any:
        """Try to find the paper channel by display name."""
        teamapi = self._mattermost.teams
        channelapi = self._mattermost.channels

        for team in teamapi.get_user_teams("me"):
            channels = [channel["id"] for channel
                        in channelapi.get_channels_for_user("me", team["id"])
                        if channel["display_name"] == channelname]
            if channels:
                # lets just hope no-one has the same channel name in multiple
                # teams
                self.teamname = team["name"]
                return channels[0]

        print(f"Channel {channelname} does not exits")
        raise ConfigError

    def _get_reporter(self: "Mattermost", userid: str) -> str:
        """Load user from mattermost api and cache."""
        userapi = self._mattermost.users
        if userid not in self._reporters:
            self._reporters[userid] = userapi.get_user(userid)["username"]

        return self._reporters[userid]

    def _get_posts_for_channel(self: "Mattermost",
                               since: Optional[int]) -> List[Dict[str, Any]]:
        """Get posts from mattermost and concat paged response since date."""
        params = {"since": since} if since else {}
        posts = []

        while True:
            resp = self._mattermost.posts.get_posts_for_channel(
                self._channel, params)
            posts.extend(resp['posts'].values())
            if resp["prev_post_id"]:
                params["before"] = resp["prev_post_id"]
            else:
                break

        return posts

    def _retrieve_all_posts(self: "Mattermost",
                            since: Optional[int]) -> List[PostDTO]:
        """Retrieve all posts from mattermost, unfiltered for papers."""
        start = time.perf_counter()
        posts = self._get_posts_for_channel(since)
        dtos: List[PostDTO] = []

        # follow children for each paper post and
        # try to extract meta information
        if config.debug:
            with open("all_posts.json", "w") as filehandle:
                json.dump(posts, filehandle, indent=4)

        for post in posts:
            doi = Doi().extract_doi(post['message'])
            if doi is not None:
                dtos.append(PostDTO(
                            id=post['id'],
                            create_at=post['create_at'],
                            message=post['message'],
                            reporter=self._get_reporter(post['user_id']),
                            link=(f"https://{self._url}/"
                                  f"{self.teamname}/pl/{post['id']}"),
                            doi=doi,))

        for dto in dtos:
            res = extract_summary_keywords(dto.message)
            dto.message = res['message']
            dto.summary += res['summary']
            dto.keywords.extend(res['keywords'])

            children = [p["message"] for p in posts if p["root_id"] == dto.id]
            for child in children:
                res = extract_summary_keywords(child)
                dto.summary += res['summary']
                dto.keywords.extend(res['keywords'])

        logging.debug("retrieving and processing %i mattermost posts took %f",
                      len(posts), time.perf_counter() - start)

        return dtos

    # def is_paperpost(self: "Mattermost", post: Dict[str, Any]) -> bool:
    #     """Check if post contains a paper reference."""
    #     return Doi().extract_doi(post['message']) is not None

    def retrieve(self: "Mattermost",
                 since: Optional[int] = None) -> List[PostDTO]:
        """Retrieve papers from mattermost channel."""
        if not self._loggedin:
            self.login()
        self.posts = self._retrieve_all_posts(since)
        return self.posts

    def check_doi_exits(self: "Mattermost", doi: str) -> bool:
        """Check for doi in current paper list."""
        doi_needle = Doi().extract_doi(doi)
        posts_found = [post for post in self.posts if post.doi == doi_needle]
        return bool(posts_found)

    def post(self: "Mattermost", message: str) -> None:
        """Post message to thread."""
        self._mattermost.posts.create_post({"channel_id": self._channel,
                                            "message": message})



"""Bibtex tests."""
from typing import Any
import pytest
from papersurfer.bibtex import Bibtex


@pytest.mark.skip(reason="needs network")  # type: ignore
def test_entry_from_doi(shared_datadir: Any) -> None:
    """Load bibtex data for doi."""
    dois = ["10.1016/j.chemgeo.2020.119702", "10.1029/2019jb019149"]

    for doi in dois:
        # FIXME: request needs to be mocked pylint # pylint: disable=fixme
        #  requests_mock.get(f"", text="")
        entry = Bibtex().entry_from_doi(doi)

        filename = doi.replace("/", "__") + ".bibtex"
        with open(shared_datadir / filename, "r") as file:
            assert entry == file.read()



"""Test parts of the ui."""
from papersurfer.ui.paperdetail import PaperDetail
from papersurfer.dtos import PostDTO


def details2text(details: PaperDetail) -> str:
    """Render details into string for testing."""
    return " ".join([e[0].text for e in details.render().contents])


def test_details() -> None:
    """Generate paper details popup."""
    details = PaperDetail()
    details.render()
    post = PostDTO(message="MSG", summary="summary")
    details.post = post
    assert details.post == post
#    details.paper = PaperDTO(title="Paper Title")
    text = details2text(details)
    assert "summary" in text



"""Paper surfer - browse papers posted on the mattermost channel.

UI:

[____(filter)______]
1. paper (open discussion) (open paper)
2. paper (open discussion) (open paper)
3. paper (open discussion) (open paper)
4. paper (open discussion) (open paper)

"""
import time
import os
import sys
import logging
import pathlib
from typing import Any, List
import configargparse  # type: ignore
from xdgappdirs import AppDirs  # type: ignore
from .exceptions import ConfigError
from .mattermost import Mattermost
from .bibtex import Bibtex
from . import config
from .papersurfer import Papersurfer
from .ui.papersurferui import PapersurferUi


def get_config_file_paths() -> List[str]:
    """Find config file paths.

    The first config file that is found is used, it is searched for (in
    this order), at:
     - config, if set (e.g. from the cli)
     - from the default source path (./papersurfer.conf)
     - home config path (system-dependent,
        see https://pypi.org/project/xdgappdirs/ for details)
     - system path (system-dependent,
        see https://pypi.org/project/xdgappdirs/ for details)

    >>> type(get_config_file_paths())
    <class 'list'>
    """
    appdir = AppDirs("papersurfer")
    paths = [appdir.site_config_dir, appdir.user_config_dir, "./"]
    default_filename = "papersurfer.conf"

    return [os.path.join(p, default_filename) for p in paths]


def get_log_file_path() -> str:
    """Get log file path."""
    path = AppDirs("papersurfer").user_log_dir
    default_filename = "papersurfer.log"

    return os.path.join(path, default_filename)


def interactive_configuration() -> Any:
    """Query user for credentials."""
    url = input("Mattermost URL (eg. mattermost.example.net): ")
    url = url.replace("https://", "").strip('/')  # cleanup https://mtm/ urls
    channel = input("Channel (eg. Paper Club): ")
    username = input("Username (same as mattermost login, "
                     "eg. JohnDoe@example.net): ")
    password = input("Password (same as mattermost login, eg. SuperSecret1): ")
    return url, channel, username, password


def get_version() -> str:
    """Get version number from static version text file."""
    pkgbase = pathlib.Path(__file__).parent
    with open(f"{pkgbase}/_version.txt", "r") as versionfile:
        return versionfile.read()


def print_paths() -> None:
    """Get paths of config, log and data files."""
    paths = get_config_file_paths()
    paths.append(get_log_file_path())
    paths.append(config.datadir)
    paths = [p for p in paths if os.path.exists(p)]
    print("\n".join(paths))


def parse_args() -> Any:
    """Parse command line arguments and config file."""
    parser = configargparse.ArgParser(
        default_config_files=get_config_file_paths())
    parser.add("-w", "--write-out-config-file",
               help="takes the current command line args and writes them out "
                    "to a config file at the given path",
               is_write_out_config_file_arg=True)
    parser.add('-c', '--config', required=False, is_config_file=True,
               help='config file path')
    parser.add('--url', required=False, help='Mattermost url')
    parser.add('--channel', required=False, help='Mattermost channel')
    parser.add('-u', '--username', required=False, help='Mattermost username')
    parser.add('-p', '--password', required=False, help='Mattermost password')
    parser.add('--dump-posts', action='store_true',
               help="Dump mattermost paper posts to stdout and exit")
    parser.add('--dump-bibtex', action='store_true',
               help="Dump bibtex collection to stdout and exit")
    parser.add('--version', action='version', version=get_version())
    parser.add("--debug", action="store_true",
               help="Print debug messages and dump debug data files")
    parser.add('--paths', action="store_true",
               help="Print config and data paths and exit")
    parser.add('--datadir', default=AppDirs("papersurfer").user_data_dir,
               help='Path to datafiles.')

    options = parser.parse_args()

    config.debug = options.debug

    pathlib.Path(options.datadir).mkdir(parents=True, exist_ok=True)
    config.datadir = options.datadir

    if options.paths:
        print_paths()
        sys.exit(0)

    if not options.url:
        start_interactive = input(
            "Could not load config file or read command line arguments, do you"
            " wish to start the interactive configuration assistant? (y/n) ")
        if start_interactive == "y":
            url, channel, username, password = interactive_configuration()
            try:
                Mattermost(url, channel, username, password)
            except ConfigError:
                print("Failed to validate configuration, exiting.")
                sys.exit(1)

            options.url = url
            options.channel = channel
            options.username = username
            options.password = password
            configpath = AppDirs("papersurfer").user_config_dir

            pathlib.Path(configpath).mkdir(parents=True, exist_ok=True)

            configfile = os.path.join(configpath, "papersurfer.conf")
            with open(configfile, "w") as file:
                file.write(f"url = {url}\n")
                file.write(f"channel = {channel}\n")
                file.write(f"username = {username}\n")
                file.write(f"password = {password}\n")
                print(f"Configfile {configfile} written.")

            time.sleep(2)
        else:
            parser.print_help()
            sys.exit(0)

    return options


def just_posts(url: str, channel: str, username: str, password: str) -> None:
    """Fuck off with all this interactive shit."""
    posts = Papersurfer(url, channel, username, password).get_posts()
    for post in posts:
        print(post.message)


def just_bibtex(url: str, channel: str, username: str, password: str) -> None:
    """Retrieve and dump bibtext formated data, unfiltered."""
    posts = Papersurfer(url, channel, username, password).get_posts()
    dois = [post.doi for post in posts]
    print(Bibtex().bib_from_dois(dois))


def main() -> None:
    """Run main program."""
    opt = parse_args()

    logfile = get_log_file_path()
    pathlib.Path(os.path.dirname(logfile)).mkdir(parents=True, exist_ok=True)
    logging.basicConfig(filename=logfile, level=logging.DEBUG)

    if opt.dump_posts:
        just_posts(opt.url, opt.channel, opt.username, opt.password)
    elif opt.dump_bibtex:
        just_bibtex(opt.url, opt.channel, opt.username, opt.password)
    else:
        PapersurferUi(opt.url, opt.channel, opt.username, opt.password)


if __name__ == "__main__":
    main()



"""Simplified DOI interface."""
from typing import List
from .doi import Doi


class Bibtex:
    """Interface for bibtex string."""
    def entry_from_doi(self: 'Bibtex', doi: str) -> str:
        """Get bibtex string for doi."""
        return Doi().get_bibtex(doi)

    def bib_from_dois(self: 'Bibtex', dois: List[str]) -> str:
        """Get bibtex string for mulitple dois."""
        return "\n".join([Doi().get_bibtex(doi) for doi in dois])



"""DOI tests."""
from typing import Any
from papersurfer.doi import Doi
from papersurfer.dtos import PaperDTO


def test_parse_doi_json(shared_datadir: Any) -> None:
    """Parse paper info."""
    datafiles = ['data_wo_author.json', 'doi-10.10292011jb008747.json',
                 'doi-wo-given-name.json', 'doi-wo-family-name.json']
    for datafile in datafiles:
        data = (shared_datadir / datafile).read_text(encoding='utf-8')
        doi = Doi()
        paper = doi.parse_doi_json(data)
        assert isinstance(paper, PaperDTO)
        assert paper.author
        assert paper.authors
        assert paper.title
        assert paper.journal
        assert paper.year
        assert paper.abstract
        assert paper.doi
        assert paper.slug
        if datafile == "data_wo_author.json":
            assert paper.slug == "N/A"


def test_get_info() -> None:
    """Return None for missing doi data."""
    Doi.load_doi_data = lambda _self, _doi: ""  # type: ignore
    info = Doi().get_info("dummy-doi")
    assert info is None



"""Tests for command line interface (CLI)."""
import os
from typing import Any
import pytest
import papersurfer


def test_runas_module() -> None:
    """Run package as a Python module."""
    exit_status = os.system('python -m papersurfer --help')
    assert exit_status == 0


def test_entrypoint() -> None:
    """Is entrypoint script installed? (setup.py)."""
    exit_status = os.system('papersurfer --help')
    assert exit_status == 0


def test_configfile_fail() -> None:
    """Fail when trying to load a nonexistent config file."""
    cli = 'python -m papersurfer --config /dev/null'
    exit_status = os.system(cli)
    assert exit_status != 0


def test_configfile(shared_datadir: Any) -> None:
    """Load and write a config file."""
    conffile = (shared_datadir / "papersurfer.conf")
    outfile = (shared_datadir / "delme.txt")
    cli = f"python -m papersurfer --config {conffile} -w {outfile}"

    exit_status = os.system(cli)
    assert exit_status == 0

    with open(outfile) as file:
        content = file.read()

    assert "Paperclub" in content


def test_cli() -> None:
    """Test if CLI stops execution w/o a command argument."""
    with pytest.raises(SystemExit):
        papersurfer.__main__.main()  # type: ignore
        pytest.fail("CLI doesn't abort asking for a command argument")



"""Dialog to add another post to public list."""
from typing import Any, Callable
import urwid  # type: ignore
from papersurfer.mattermost import Mattermost
from .prettybutton import PrettyButton
from ..doi import Doi
from ..dtos import PaperDTO


class PostDialog(urwid.WidgetWrap):  # type: ignore
    """Dialog to submit a new paper to mattermost thread.

    UI:
        DOI: [ _________________]
        Generated Message:
            "# # # #  # # # #"

        [Submit]       [Close]
    """
    def __init__(self: "PostDialog", mattermost: Mattermost,
                 close: Callable[[Any], Any], loop: Any) -> None:
        self._loop = loop
        self.alarm_handle = None

        self.doi: str = ""
        self.msg: str = ""
        self.mattermost = mattermost
        self.close = close
        self.doi_input = urwid.Edit("Doi: ")
        urwid.connect_signal(self.doi_input, 'change', self.h_input)
        self.doi_result = urwid.Text("")

        body_pile = urwid.Pile([
            self.doi_input,
            urwid.Divider(" "),
            self.doi_result,
            urwid.Divider(" "),
            urwid.Columns([
                PrettyButton("Close", self.close),
                PrettyButton("Submit", self.submit)
            ]),
        ])
        body_filler = urwid.Filler(body_pile, valign='top')
        body_padding = urwid.Padding(body_filler, left=1, right=1)
        body = urwid.LineBox(body_padding)
        frame = urwid.Frame(body,
                            header=urwid.Text("Submit new paper to list"))

        self.widget = frame

        super().__init__(self.widget)

    def submit(self: "PostDialog", _: Any) -> None:
        """Submit post to thread."""
        if not self.mattermost.check_doi_exits(self.doi):
            self.mattermost.post(self.msg)
        self.close(_)

    def create_mgs(self: "PostDialog", paper: PaperDTO) -> str:
        """Format post message."""
        msg = f"""\
{paper.title}
{paper.authors}
{paper.journal} [{paper.slug}]
{Doi().get_doi_link(paper.doi)}"""
        return msg

    def h_input(self: "PostDialog", _: Any, doi: str) -> None:
        """Handle doi input field and debounce."""
        self.doi_result.set_text("")
        self.doi = ""
        self.msg = ""

        self._loop.remove_alarm(self.alarm_handle)
        self.alarm_handle = self._loop.set_alarm_in(.5, self.search_doi, doi)

    def search_doi(self: "PostDialog", _loop: Any, doi: str) -> None:
        """Trigger search for paper ref by doi and update ui."""
        self.doi_result.set_text("... loading ...")
        self._loop.draw_screen()
        if Doi().extract_doi(doi):
            paper = Doi().get_info(doi)
            if paper:
                if self.mattermost.check_doi_exits(doi):
                    self.doi_result.set_text(f"{self.create_mgs(paper)} \n"
                                             "-> Paper already posted! <-")
                else:
                    self.doi_result.set_text(self.create_mgs(paper))
                    self.doi = doi
                    self.msg = self.create_mgs(paper)
            else:
                self.doi_result.set_text("Doi not found.")
            return

        self.doi_result.set_text("invalid doi")



"""Highlight string in strings."""
import re
import urwid  # type: ignore


def highlight(text: str, needle: str) -> urwid.Text:
    """Highlight a search term in a string."""
    text_items = []
    needle = needle or "ßß"
    needles = re.findall(needle, text, flags=re.IGNORECASE)
    hay = re.split(needle, text, flags=re.IGNORECASE)
    for i, item in enumerate(hay):
        text_items.append(('', item))
        if i < len(needles):
            text_items.append(('needle', needles[i]))

    return text_items



"""Papersurfer main interface."""
import os
from typing import Any, Dict, List, Optional
from tinydb import TinyDB, Query
from .mattermost import Mattermost
from .bibtex import Bibtex
from .dtos import PostDTO, PaperDTO
from . import config


class Papersurfer:
    """Organize and cache paper/post data.

    This handles interaction with mattermost, doi and a local database.

    Papers and posts are similar but distinct concepts. A post contains
    information on a single mattermost entry, containing a paper reference.
    A paper contains information on a single scientific paper and a reference
    back to the mattermost post.
    """
    def __init__(self: "Papersurfer", url: str, channelname: str,
                 username: str, password: str) -> None:
        self.searchstring = ""

        self.mattermost = Mattermost(url, channelname, username, password)

        self.db_path = config.datadir

        self.db_posts: Any = None
        self.db_papers: Any = None
        self.db_files = {
            "posts": "papersurfer_posts_db.json",
            "papers": "papersurfer_papers_db.json",
        }
        self._posts: List[PostDTO] = []

    def load(self: "Papersurfer") -> None:
        """Load data from mattermost and save to storage."""
        self._connect_db()
        latest = self.get_latest_post()
        posts = self.mattermost.retrieve(latest["create_at"]
                                         if latest else None)
        self._update_db(posts=posts)

    def _connect_db(self: "Papersurfer") -> None:
        """Establish db connection. Noop if already connected."""
        if not self.db_posts:
            dbfile = os.path.join(self.db_path, self.db_files['posts'])
            self.db_posts = TinyDB(dbfile)  # type: ignore
        if not self.db_papers:
            dbfile = os.path.join(self.db_path, self.db_files['papers'])
            self.db_papers = TinyDB(dbfile)  # type: ignore

    def get_latest_post(self: "Papersurfer") -> Optional[Dict[str, Any]]:
        """Find the newest post and return."""
        posts = self.db_posts.all()
        if posts:
            posts.sort(reverse=True, key=lambda p: p["create_at"])

        return posts[0] if posts else None

    def _update_db(self: "Papersurfer", posts: Optional[List[PostDTO]] = None,
                   papers: Optional[List[PaperDTO]] = None) -> None:
        """Merge new data into database."""
        if posts:
            self._upsert_multiple(posts, self.db_posts)
        if papers:
            self._upsert_multiple(papers, self.db_papers)

    def _upsert_multiple(self: "Papersurfer", records: Any,
                         database: Any) -> None:
        """Update record in db unless it exits, then insert.

        Would be trivial if we could just change the unique id in tinydb to the
        doi property, but we can't.
        """
        for record in records:
            database.upsert(record.__dict__,
                            Query().doi == record.doi)  # type: ignore

    def _db_entry2post(self: "Papersurfer", entry: Dict[str, Any]) -> PostDTO:
        """Convert database entry to PostDTO."""
        return PostDTO(
            id=entry["id"],
            create_at=entry["create_at"],
            message=entry["message"],
            reporter=entry["reporter"],
            doi=entry["doi"],
            link=entry["link"] if "link" in entry else "",
            keywords=entry["keywords"] if "keywords" in entry else [],
            summary=entry["summary"] if "summary" in entry else "")

    def get_posts(self: "Papersurfer") -> List[PostDTO]:
        """Get all posts in storage."""
        if not self._posts:
            self._connect_db()
            self._posts = [self._db_entry2post(p) for p in self.db_posts.all()]
        return self._posts

    def filter_post(self: "Papersurfer", post: PostDTO,
                    searchstring: str) -> bool:
        """Determine if needle is in post."""
        searchstring = searchstring.lower()
        message = post.message.lower()
        reporter = post.reporter.lower()
        summary = post.summary.lower()
        keywords = (",".join(post.keywords)).lower()
        return (searchstring in message or searchstring in reporter
                or searchstring in summary or searchstring in keywords)

    def get_posts_filtered(self: "Papersurfer") -> List[PostDTO]:
        """Return a list of papers, filtered by filter."""
        return [post for post in self.get_posts()
                if self.filter_post(post, self.searchstring)]

    def get_papers(self: "Papersurfer") -> Any:
        """Get all papers in storage."""
        return self.db_papers.all()

    def export_to_bibtex(self: "Papersurfer") -> None:
        """Export current filtered list to bibtex file."""
        papers = self.get_posts_filtered()
        dois = [paper.doi for paper in papers]
        string = Bibtex().bib_from_dois(dois)
        with open("export.bib", 'w') as file:
            file.write(string)



import pathlib
from setuptools import setup, find_packages  # type: ignore

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()
VERSION = (HERE / "papersurfer/_version.txt").read_text()

# This call to setup() does all the work
setup(
    name="papersurfer",
    version=VERSION,
    description="",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Johann Jacobsohn",
    author_email="johann.jacobsohn@uni-hamburg.de",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        'Development Status :: 1 - Planning',
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "requests",
        "mattermostdriver",
        "urwid",
        "configargparse",
        "tinydb",
        "dataclasses",
        "xdgappdirs",
    ],
    extras_require={
        'dev': [
            "tox",
            "flake8",
            "flake8-annotations",
            "pylint>=2.6.0",
            "pycodestyle",
            "pydocstyle",
            "twine",
            "pytest",
            "pytest-cov",
            "pytest-datadir",
            "pytest-mock",
            "mypy",
        ]
    },
    entry_points={
        "console_scripts": [
            "papersurfer=papersurfer.__main__:main",
        ]
    },
)



"""Tests for mattermost adapter."""
import json
from typing import Any
from papersurfer.mattermost import Mattermost
from papersurfer.dtos import PostDTO


def test_get_posts_with_keywords_and_summaries(mocker: Any,
                                               shared_datadir: Any) -> None:
    """Check if posts can be loaded and parsed."""
    url = ""
    channelname = ""
    username = ""
    password = ""

    mattermost = Mattermost(url, channelname, username, password)

    with open((shared_datadir / "posts_keywords.json")) as fhandle:
        res = json.load(fhandle)

    mocker.patch('papersurfer.mattermost.Mattermost._get_reporter',
                 return_value="Name")
    mocker.patch('papersurfer.mattermost.Mattermost.login',
                 return_value=True)
    mocker.patch('papersurfer.mattermost.Mattermost._get_posts_for_channel',
                 return_value=res)

    posts = mattermost.retrieve()

    assert len(posts) == 3
    assert isinstance(posts[2], PostDTO)

    assert "Probabilistic hazard maps for operational use" in posts[1].message
    assert "Holuhraun" in posts[1].keywords
    assert "The Holuhraun fissure eruption" in posts[1].summary

    assert "Mantle and crustal sources" in posts[2].message
    assert not posts[2].keywords
    assert "We present a new high‐resolution seismic model" in posts[2].summary

    assert "Automatic detection of volcanic surface" in posts[0].message
    assert not posts[0].keywords
    assert "Novel application of CNN is capable" in posts[0].summary



"""Data transfer objects."""
# pylint: disable=too-many-instance-attributes
from dataclasses import dataclass, field
from typing import List


@dataclass
class PostDTO:
    """Encapsulate Mattermost Posts."""
    id: str = ""
    link: str = ""
    create_at: int = -1
    message: str = ""
    reporter: str = ""
    doi: str = ""
    keywords: List[str] = field(default_factory=list)
    summary: str = ""

    def __str__(self: "PostDTO") -> str:
        return self.message + '\n' + str(self.keywords) + '\n' + self.summary


@dataclass
class PaperDTO:
    """Encapsulate Paper meta data."""
    author: str = ""
    authors: str = ""
    title: str = ""
    journal: str = ""
    year: int = 0
    abstract: str = ""
    doi: str = ""
    slug: str = ""
